from pathlib import Path
from platformdirs import PlatformDirs

dirs = PlatformDirs("Shoggoth", "Shoggoth")
root_dir = Path(dirs.user_data_dir)

asset_dir = root_dir / "assets"
defaults_dir = asset_dir / "defaults"
font_dir = asset_dir / "fonts"
overlay_dir = asset_dir / "overlays"
template_dir = asset_dir / "templates"
icon_dir = asset_dir / "icons"



import os
from time import time
from kivy.app import App
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.label import Label
from kivy.core.window import Window
from kivy.properties import StringProperty
from kivy.clock import Clock
from kivy.lang import Builder

from shoggoth.file_monitor import FileMonitor
from shoggoth.project import Project
from shoggoth.renderer import CardRenderer
from shoggoth.files import defaults_dir
import json


class ViewerRoot(BoxLayout):
    """Root widget for the Viewer mode"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        Window.bind(on_key_down=self.on_keyboard)
        self.shown = 'both'
        self._back_container = self.ids.back_container
        self._front_container = self.ids.front_container

    def toggle_shown(self, target='both'):
        if target == 'front':
            if not self._front_container.parent:
                self.ids.master_container.add_widget(self._front_container)
            if self._back_container.parent:
                self.ids.master_container.remove_widget(self._back_container)
            self.shown = 'front'
        if target == 'back':
            if not self._back_container.parent:
                self.ids.master_container.add_widget(self._back_container)
            if self._front_container.parent:
                self.ids.master_container.remove_widget(self._front_container)
            self.shown = 'back'
        if target == 'both':
            if not self._front_container.parent:
                self.ids.master_container.add_widget(self._front_container)
            if not self._back_container.parent:
                self.ids.master_container.add_widget(self._back_container)
            self.shown = 'both'


    def on_keyboard(self, instance, keyboard, keycode, text, modifiers):
        print('on keyboard', keycode, modifiers, text)
        # Handle keyboard shortcuts
        if keycode == 27:  # Esc key
            App.get_running_app().stop()
            return True
        if keycode == 80:  # left
            App.get_running_app().select_card(-1)
            return True
        if keycode == 79:  # right
            App.get_running_app().select_card(1)
            return True
        if text == 'n':  # right
            App.get_running_app().number_cards()
            return True
        if text == 's' and 'ctrl' in modifiers:  # right
            App.get_running_app().save_all()
            return True
        if text == 'f':  # right
            if 'shift' in modifiers:
                App.get_running_app().toggle_faces_shown()
                return True
            App.get_running_app().flip_card()
            return True
        if text == 'e':
            if modifiers == ['ctrl']:
                f = App.get_running_app().export_all
                Clock.schedule_once(lambda x: f())
            else:
                f = App.get_running_app().export_current
                Clock.schedule_once(lambda x: f())
        return False

class ViewerApp(App):
    """Application class for Viewer Mode"""
    file_path = StringProperty("")
    status_message = StringProperty("Ready")

    def __init__(self, file_path, **kwargs):
        super().__init__(**kwargs)
        self.file_path = file_path

    def build(self):
        # Set window title and size
        self.icon = 'assets/elder_sign_neon.ico'
        self.title = f"Shoggoth - {os.path.basename(self.file_path)}"

        # Initialize the root widget
        self.root = ViewerRoot()
        self.data = {}

        # Create a status bar
        self.status_bar = Label(
            text=self.status_message,
            size_hint_y=None,
            height=24,
            halign='left',
            text_size=(Window.width, 24),
            padding=(10, 0),
            valign='middle',
            color=(1, 1, 1, 1)
        )
        self.root.add_widget(self.status_bar)

        # Initialize card renderer
        self.card_renderer = CardRenderer()
        self.card_index = 0

        # Initialize file monitoring
        self.file_monitor = FileMonitor(os.path.dirname(self.file_path), self.on_file_changed)
        self.file_monitor.add_file(self.file_path)
        self.file_monitor.start()
        self.file_monitor_settings = FileMonitor(os.path.dirname(defaults_dir), self.on_file_changed)
        self.file_monitor_settings.start()

        # Bind property changes
        self.bind(status_message=self._update_status_bar)

        # Load the initial card
        Clock.schedule_once(lambda dt: self.load_file(self.file_path), 0.5)

        return self.root

    def _update_status_bar(self, instance, value):
        """Update the status bar text"""
        self.status_bar.text = value

    def select_card(self, increment):
        if not self.cards:
            return
        self.card_index += increment
        self.card_index = self.card_index % len(self.cards)
        self.render_card(self.cards[self.card_index], None)
        self.status_message = 'Moved ' + ('left' if increment == -1 else 'right')

    def on_file_changed(self, file_path):
        """Handle file changes"""
        if file_path == self.file_path:
            self.load_file(file_path)
        else:
            self.load_file(self.file_path)

    def render_card(self, card, file_path=None):
        # Update card preview
        front_image, back_image = self.card_renderer.get_card_textures(card)
        target = self.root.ids.front_preview
        target.texture = front_image
        self.root.ids.back_preview.texture = back_image

        # Update status
        if file_path:
            self.status_message = f"Loaded: {os.path.basename(file_path)} (Last modified: {os.path.getmtime(file_path)})"

    def number_cards(self):
        self.cards[self.card_index].expansion.assign_card_numbers()
        self.status_message = f"Updated card numbers {time()}"

    def load_file(self, file_path):
        """Load a card from file and update the UI"""
        if not file_path or not os.path.exists(file_path):
            self.status_message = f"File not found: {file_path}"
            return

        try:
            # Check if file is a card file
            with open(file_path, 'r') as file:
                new_data = json.load(file)

            # look for card diffs
            if self.data:
                cards = Project(file_path, new_data).get_all_cards()
                for card in cards:
                    if card not in self.cards:
                        self.card_index = cards.index(card)
                        break
                self.cards = cards
            else:
                self.data = new_data
                self.cards = Project(file_path, new_data).get_all_cards()

            Clock.schedule_once(lambda x: self.render_card(self.cards[self.card_index], file_path))

        except Exception as e:
            self.status_message = f"Error loading file: {str(e)}"

    def save_all(self):
        self.cards[self.card_index].expansion.save()

    def on_stop(self):
        """Clean up when the application stops"""
        if self.file_monitor:
            self.file_monitor.stop()

    def export_current(self):
        export_folder = os.path.join(os.path.dirname(self.file_path), f'export_of_{os.path.basename(self.file_path).split(".")[1]}')
        if not os.path.exists(export_folder):
            os.mkdir(export_folder)
        t = time()
        card = self.cards[self.card_index]
        self.card_renderer.export_card_images(card, export_folder)
        print(f'Export of {card.name} cards done in {time()-t} seconds')

    def export_all(self):
        export_folder = os.path.join(os.path.dirname(self.file_path), f'export_of_{os.path.basename(self.file_path).split(".")[1]}')
        if not os.path.exists(export_folder):
            os.mkdir(export_folder)
        t = time()
        for card in self.cards:
            self.card_renderer.export_card_images(card, export_folder)
        print(f'Export of {len(self.cards)} cards done in {time()-t} seconds')

    def toggle_faces_shown(self):
        if self.root.shown == 'both':
            self.root.toggle_shown('front')
        else:
            self.root.toggle_shown('both')
        print(self.root.shown)

    def flip_card(self):
        self.root.toggle_shown('back' if self.root.shown == 'front' else 'front')
        print(self.root.shown)



import os
import json
from uuid import uuid4
from pathlib import Path
from typing import Any, Dict, Optional

from kivy.app import App

from shoggoth.files import defaults_dir


class Face:
    def __init__(self, data, card=None, encounter=None, expansion=None):
        self.data = data
        self.card = card
        self.encounter = encounter
        self.expansion = expansion
        self._fallback = None

    def __eq__(self, other):
        return self.data == other.data

    @property
    def fallback(self):
        if self._fallback != None:
            return self._fallback

        if Path(self.data['type']).is_file():
            defaults_path = Path(self.data['type'])
        else:
            defaults_file = f'{self.data["type"]}.json'
            defaults_path = defaults_dir / defaults_file

        try:
            with defaults_path.open('r') as f:
                self._fallback = json.load(f)
            return self._fallback
        except Exception as e:
            print('exception when loading fallback:', e)
            return {}

    def __getitem__(self, key):
        if key in self.data:
            return self.data[key]
        return self.fallback[key]

    def get(self, key, default=''):
        if key in self.data:
            return self.data[key]
        if key in self.fallback:
            return self.fallback[key]
        else:
            return default

    def set(self, key, value):
        if key == 'type':
            self._fallback = None
        self.data[key] = value
        App.get_running_app().update_card_preview()


class Card:
    """Class to represent the card object and file structure"""

    def __init__(
        self,
        data:Dict[str, Any],
        expansion,
        encounter=None,
    ):
        self.app = None
        self.data = data
        self.name = data['name']
        self.encounter = encounter
        self.expansion = expansion
        if 'id' not in data:
            data['id'] = str(uuid4())
        self.id = data['id']

        self.front = Face(self.data['front'], card=self)
        self.back = Face(self.data['back'], card=self)
        self.app = App.get_running_app()

    @property
    def expansion_number(self):
        return self.data.get('expansion_number', -1)

    @expansion_number.setter
    def expansion_number(self, value):
        self.data['expansion_number'] = value

    @property
    def encounter_number(self):
        if not self.encounter:
            return None
        return self.data.get('encounter_number', -1)

    @encounter_number.setter
    def encounter_number(self, value):
        if not self.encounter:
            raise Exception("No encounter, can't set number.")
        self.data['encounter_number'] = value

    @property
    def code(self):
        if self.encounter:
            return f'{self.expansion.code}_{self.encounter.code}_{self.name}'
        return f'{self.expansion.code}_{self.expansion_number}_{self.name}'

    def __eq__(self, other):
        return self.data == other.data

    def get_class(self):
        back_classes = self.data['back'].get('classes')
        classes = self.data['front'].get('classes', back_classes)
        if not classes:
            return None

        if len(classes) > 1:
            return 'multi'
        else:
            return classes[0]


    @staticmethod
    def is_valid(data):
        return 'front' in data and 'back' in data

    def save(self):
        """Save card data to JSON file"""
        pass

    @staticmethod
    def new(name) -> Dict[str, Any]:
        return {
            'name': name,
            'id': str(uuid4()),
            'front': {
                'type': 'asset'
            },
            'back': {
                'type': 'player'
            }
        }

# templates
class TEMPLATES:
    @classmethod
    def get(cls, name):
        return {
            'investigator': cls.INVESTIGATOR(),
            'asset': cls.ASSET(),
            'event': cls.EVENT(),
            'skill': cls.SKILL(),
            'story': cls.STORY(),
            'treachery': cls.TREACHERY(),
            'enemy': cls.ENEMY(),
            'location': cls.LOCATION(),
            'act': cls.ACT(),
            'agenda': cls.AGENDA(),
            'scenario': cls.SCENARIO(),
        }.get(name, cls.BASE())

    @classmethod
    def BASE(cls):
        return {
            'name': '',
            'id': str(uuid4()),
            'front': {
                'type': ''
            },
            'back': {
                'type': ''
            }
        }

    @classmethod
    def ASSET(cls):
        card = cls.BASE()
        card['front']['type'] = 'asset'
        card['back']['type'] = 'player'
        return card

    @classmethod
    def INVESTIGATOR(cls):
        card = cls.BASE()
        card['front']['type'] = 'investigator'
        card['back']['type'] = 'investigator_back'
        return card

    @classmethod
    def EVENT(cls):
        card = cls.BASE()
        card['front']['type'] = 'event'
        card['back']['type'] = 'player'
        return card

    @classmethod
    def SKILL(cls):
        card = cls.BASE()
        card['front']['type'] = 'skill'
        card['back']['type'] = 'player'
        return card

    @classmethod
    def ENEMY(cls):
        card = cls.BASE()
        card['front']['type'] = 'enemy'
        card['back']['type'] = 'encounter'
        return card

    @classmethod
    def TREACHERY(cls):
        card = cls.BASE()
        card['front']['type'] = 'treachery'
        card['back']['type'] = 'encounter'
        return card

    @classmethod
    def LOCATION(cls):
        card = cls.BASE()
        card['front']['type'] = 'location'
        card['back']['type'] = 'location_back'
        return card

    @classmethod
    def ACT(cls):
        card = cls.BASE()
        card['front']['type'] = 'act'
        card['back']['type'] = 'act_back'
        return card

    @classmethod
    def AGENDA(cls):
        card = cls.BASE()
        card['front']['type'] = 'agenda'
        card['back']['type'] = 'agenda_back'
        return card

    @classmethod
    def SCENARIO(cls):
        card = cls.BASE()
        card['front']['type'] = 'chaos'
        card['back']['type'] = 'chaos_back'
        return card

    @classmethod
    def STORY(cls):
        card = cls.BASE()
        card['front']['type'] = 'story'
        card['back']['type'] = 'story_back'
        return card



import os
import time
from threading import Thread
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler
import logging
#logging.getLogger('watchdog').setLevel(logging.WARNING)


class CardFileHandler(FileSystemEventHandler):
    """File system event handler for card files"""

    def __init__(self, callback, monitored_files=None):
        super(CardFileHandler, self).__init__()
        self.callback = callback
        self.monitored_files = monitored_files or []
        self.last_modified = {}

    def on_modified(self, event):
        """When a file is modified"""
        if event.is_directory:
            return

        if not self.monitored_files or event.src_path in self.monitored_files:
            # Add a small delay to avoid multiple reload triggers
            current_time = time.time()
            last_time = self.last_modified.get(event.src_path, 0)

            if current_time - last_time > 0.5:
                self.last_modified[event.src_path] = current_time
                self.callback(event.src_path)

class FileMonitor:
    """Monitors files for changes"""

    def __init__(self, project_path, callback):
        self.project_path = project_path
        self.callback = callback
        self.observer = None
        self.monitored_files = set()

    def start(self):
        """Start monitoring files"""
        if self.observer:
            self.stop()

        self.observer = Observer()
        handler = CardFileHandler(self.callback, self.monitored_files)

        # Monitor the entire project directory
        self.observer.schedule(handler, self.project_path, recursive=True)
        self.observer.start()

    def stop(self):
        """Stop monitoring files"""
        if self.observer:
            self.observer.stop()
            self.observer.join()
            self.observer = None

    def add_file(self, file_path):
        """Add a file to monitor"""
        if os.path.exists(file_path):
            self.monitored_files.add(file_path)

    def remove_file(self, file_path):
        """Remove a file from monitoring"""
        if file_path in self.monitored_files:
            self.monitored_files.remove(file_path)



from time import time
from PIL import Image, ImageDraw, ImageFont, ImageOps
from kivy.graphics.texture import Texture
from kivy.core.image import Image as kivy_img
from kivy.uix.image import CoreImage
import sys, os
import json
import re
from io import BytesIO
from shoggoth.rich_text import RichTextRenderer
import numpy as np
from shoggoth.files import template_dir, overlay_dir, icon_dir, asset_dir, defaults_dir
from pathlib import Path

import logging
logging.getLogger('PIL').setLevel(logging.ERROR)
logging.getLogger('pillow').setLevel(logging.ERROR)


class Region:
    bleed:tuple[int, int] = (0,0)

    def __init__(self, data):
        if not data:
            data = {}
        self.x = data.get('x', 0)
        self.y = data.get('y', 0)
        if Region.bleed:
            self.x += Region.bleed[0]
            self.y += Region.bleed[1]
        self.width = data.get('width', 0)
        self.height = data.get('height', 0)

    @property
    def size(self):
        return (self.width, self.height)

    @property
    def pos(self):
        return (self.x, self.y)

    @property
    def center(self):
        return {'x': self.x + self.width // 2, 'y': self.y + self.height // 2}

    def __bool__(self):
        return self.width > 0 and self.height > 0


class CardRenderer:
    """Renders card images based on card data"""

    # Standard card dimensions
    #CARD_WIDTH = 375
    #CARD_HEIGHT = 524
    CARD_WIDTH = 750
    CARD_HEIGHT = 1050

    def __init__(self):
        # Base paths
        self.assets_path = asset_dir
        self.templates_path = template_dir
        self.overlays_path = overlay_dir
        self.icons_path = icon_dir
        self.defaults_path = defaults_dir
        self.cache = {}

        # Initialize rich text renderer
        self.rich_text = RichTextRenderer(self)

    def get_thumbnail(self, card):
        """ Renders a low res version of the front of a card """
        image = self.render_card_side(card, card.front, size=1)
        image = image.resize((int(image.width*.5), int(image.height*.5)))
        buffer = BytesIO()
        image.save(buffer, format='jpeg', quality=50)
        buffer.seek(0)
        return buffer

    def get_card_textures(self, card):
        """Render both sides of a card"""
        front, back = self.get_card_images(card)

        # Convert PIL images to Kivy textures
        from time import time
        t = time()
        front_texture = self.pil_to_texture(front)
        back_texture = self.pil_to_texture(back)
        print(f'Time to texturize: {time()-t}')

        return front_texture, back_texture

    def get_card_images(self, card):
        """Get raw PIL images for the card"""
        # Render front and back
        front_image = self.render_card_side(card, card.front)
        back_image  = self.render_card_side(card, card.back)

        return front_image, back_image

    def export_card_images(self, card, folder):
        front_image = self.render_card_side(card, card.front)
        back_image  = self.render_card_side(card, card.back)

        front_image.save(os.path.join(folder, card.code + '_front.jpeg'), quality=95)
        back_image.save(os.path.join(folder, card.code + '_back.jpeg'), quality=95)

    def pil_to_texture(self, pil_image):
        """Convert PIL image to Kivy texture"""
        buffer = BytesIO()
        pil_image.save(buffer, format='jpeg', quality=50)
        buffer.seek(0)

        # Create texture
        im = CoreImage(buffer, ext='jpeg')
        return im.texture

    def render_card_side(self, card, side, size=1):
        """Render one side of a card"""
        from time import time

        t = time()
        print(f'start render card side')
        self.current_card = card
        self.current_side = side
        self.current_opposite_side = card.front if side == card.back else card.back
        self.current_field = None

        height, width = int(self.CARD_HEIGHT*size), int(self.CARD_WIDTH*size)
        bleed = side.get('template_bleed', None)
        if bleed:
            height, width = int((self.CARD_HEIGHT+bleed[0]*2)*size), int((self.CARD_WIDTH+bleed[1]*2)*size)
            Region.bleed = (bleed[0], bleed[1])

        if side.get('orientation', 'vertical') == 'horizontal':
            card_image = Image.new('RGB', (height, width), (255, 255, 255))
        else:
            card_image = Image.new('RGB', (width, height), (255, 255, 255))

        self.render_illustration(card_image, side)
        self.render_template(card_image, side)
        if side['type'] in ('investigator'):
            self.render_illustration(card_image, side)

        for func in [
            self.render_level,
            self.render_encounter_icon,
            self.render_expansion_icon,
            self.render_act_info,
            self.render_icons,
            self.render_connection_icons,
            self.render_tokens,
            self.render_health,
            self.render_text,
            self.render_class_symbols,
            self.render_enemy_stats,
            self.render_slots,
            self.render_chaos,
            self.render_customizable,
        ]:
            try:
                func(card_image, side)
            except Exception as e:
                print(f'Failed in {func}: {e}')
        print(f'end render card side: {time()-t}')
        return card_image

    def render_text(self, card_image, side):
        for field in [
            'cost', 'name', 'traits', 'text', 'subtitle', 'label',
            'attack', 'evade', 'health', 'stamina', 'sanity',
            'clues', 'doom', 'shroud', 'willpower', 'intellect',
            'combat', 'agility', 'illustrator', 'copyright', 'collection', 'difficulty'
        ]:
            value = side.get(field)
            if not value:
                continue

            # Checkboxes - primarily for Customizable
            if field == 'text':
                cb_value = side.get('checkbox_entries')
                if cb_value:
                    lines = ''
                    for slots, name, text in cb_value:
                        lines += '\n' + '☐'*slots + f' <b>{name}.</b> {text}'
                    value += lines

            if field == 'text':
                flavor = side.get('flavor_text', '')
                if flavor:
                    value += '\n'
                    value += f'<center><i>{flavor}</i>'

            try:
                region = Region(side[f'{field}_region'])
                font = side.get(f'{field}_font', {})
                polygon = side.get(f'{field}_polygon', None)
                if font.get('rotation'):
                    alignment = font.get('alignment', 'left')
                    temp_image = Image.new('RGBA', (region.height, region.width), (255, 255, 255, 0))
                    self.rich_text.render_text(
                        temp_image,
                        value,
                        {'x': 0, 'y': 0, 'height': region.width, 'width': region.height},
                        font=font.get('font', 'regular'),
                        font_size=font.get('size', 20),
                        fill=font.get('color', '#231f20'),
                        outline=font.get('outline'),
                        outline_fill=font.get('outline_color'),
                        alignment=font.get('alignment', 'left'),
                        polygon=polygon,
                    )
                    temp_image = temp_image.rotate(-90, expand=True)
                    card_image.paste(temp_image, (region.x-10, region.y), temp_image)
                else:
                    self.rich_text.render_text(
                        card_image,
                        value,
                        region,
                        font=font.get('font', 'regular'),
                        font_size=font.get('size', 20),
                        fill=font.get('color', '#231f20'),
                        outline=font.get('outline'),
                        outline_fill=font.get('outline_color'),
                        alignment=font.get('alignment', 'left'),
                        polygon=polygon,
                    )
            except Exception as e:
                import traceback
                print(f"Error rendering field: {field}\n {e}")
                traceback.print_exc()
                continue

    def render_tokens(self, card_image, side):
        value = side.get('tokens')
        if not value:
            return

        region = Region(side['chaos_region'])
        entry_height = region.height//len(value)
        font = side.get('chaos_font', {})

        for index, icon in enumerate(value):
            text = value[icon]
            text_region = {
                'x': region.x+94,
                'y': region.y+(index*entry_height),
                'width': region.width-94,
                'height': entry_height,
            }
            self.rich_text.render_text(
                card_image,
                text,
                text_region,
                font=font.get('font', 'regular'),
                font_size=font.get('size', 20),
                fill=font.get('color', '#231f20'),
                outline=font.get('outline'),
                outline_fill=font.get('outline_color'),
                alignment=font.get('alignment', 'left')
            )

            overlay_path = self.overlays_path/f"chaos_{icon}.png"
            overlay_icon = Image.open(overlay_path).convert("RGBA")
            overlay_icon = overlay_icon.resize((overlay_icon.width*2, overlay_icon.height*2))
            card_image.paste(overlay_icon, (region.x, region.y+(index*entry_height)), overlay_icon)

    def render_connection_icons(self, card_image, side):
        # own icon
        value = side.get('connection')
        if value and value != "None":
            # ready circle for coloring
            region = side.get('connection_region')
            connection_image = Image.open(self.overlays_path/f"location_{value}.png").convert("RGBA")
            connection_image = connection_image.resize((connection_image.width*2, connection_image.height*2))
            card_image.paste(connection_image, (region.x, region.y), connection_image)

        # outgoing connections
        value = side.get('connections')
        if not value:
            return

        # grab and paint each icon
        for index, icon in enumerate(value):
            if not icon or icon == "None":
                continue
            region = Region(side.get(f'connection_{index+1}_region'))

            connection_image = Image.open(self.overlays_path/f"location_{icon}.png").convert("RGBA")
            connection_image = connection_image.resize((connection_image.width*2, connection_image.height*2))
            card_image.paste(connection_image, region.pos, connection_image)


    def render_icons(self, card_image, side):
        value = side.get('icons')
        if not value:
            return

        box_path = self.overlays_path/f"skill_box_{side['class']}.png"
        box_image = Image.open(box_path).convert("RGBA")
        box_image = box_image.resize((109, 83))
        for index, icon in enumerate(value):
            overlay_path = self.overlays_path/f"skill_icon_{icon}.png"
            overlay_icon = Image.open(overlay_path).convert("RGBA")
            overlay_icon = overlay_icon.resize((overlay_icon.width * 2, overlay_icon.height * 2))
            card_image.paste(box_image, (0, index*84+165), box_image)
            card_image.paste(overlay_icon, (25, index*84+181), overlay_icon)

    def render_act_info(self, card_image, side):
        self.current_field = 'index'
        value = side.get('index')
        region = Region(side.get('scenarioindex_region'))
        if not region or not value:
            return

        self.rich_text.render_text(card_image, value, region, alignment='center')

    def render_health(self, card_image, side):
        """ Add health and sanity overlay, if needed. """
        for stat in ['health', 'sanity']:
            self.current_field = stat
            value = side.get(stat)
            region = Region(side.get(f'{stat}_region'))
            if region is None or value is None:
                continue

            overlay_path = self.overlays_path/f"{stat}_base.png"
            overlay_icon = Image.open(overlay_path).convert("RGBA")
            overlay_icon = overlay_icon.resize((region.width, region.height))
            center_x = region.x
            center_y = region.y - 15
            card_image.paste(overlay_icon, (center_x, center_y), overlay_icon)

    def render_expansion_icon(self, card_image, side):
        """Render the encounter icon """
        region = Region(side.get('collection_portrait_clip_region'))
        if not region:
            return

        icon = side.card.expansion.icon
        if not icon:
            return

        icon = Image.open(side.card.expansion.icon, formats=['png', 'jpg']).convert('RGBA')
        if icon.width > icon.height:
            icon = icon.resize((region.width, int(region.height*(icon.height/icon.width))))
        else:
            icon = icon.resize((int(region.width*(icon.width/icon.height)), region.height))

        # invert
        alpha = icon.getchannel('A')
        icon = icon.convert('RGB')
        icon = ImageOps.invert(icon)
        icon.putalpha(alpha)
        card_image.paste(icon, (region.x, region.y), icon)

    def render_encounter_icon(self, card_image, side):
        """Render the encounter icon """
        if not side.card.encounter:
            return

        region = Region(side.get('encounter_portrait_clip_region'))
        if not region:
            raise Exception("Encounter icon region not defined")

        overlay = side.get('encounter_overlay')
        if overlay:
            overlay_image = Image.open(overlay).convert("RGBA")
            overlay_region = side.get('encounter_overlay_region')
            overlay_image = overlay_image.resize((overlay_region.width, overlay_region.height))
            card_image.paste(overlay_image, (overlay_region.x, overlay_region.y), overlay_image)

        icon = Image.open(side.card.encounter.icon).convert("RGBA")
        icon = icon.resize((region.width, region.height))
        card_image.paste(icon, (region.x, region.y), icon)

    def render_slots(self, card_image, side):
        """Render the slot icons """
        slots = side.get('slots')
        slot = side.get('slot')
        if not slots:
            slots = [slot]
        if not slots:
            return

        for index, slot in enumerate(slots):
            region = Region(side.get(f'slot_{index+1}_region', {}))
            if not region:
                continue

            path = self.overlays_path/f'slot_{slot}.png'
            slot_image = Image.open(path).convert("RGBA")
            slot_image = slot_image.resize(region.size)
            card_image.paste(slot_image, region.pos, slot_image)

    def render_enemy_stats(self, card_image, side):
        """Render Damage and horror."""
        for token in ('damage', 'horror'):
            value = int(side.get(token, '0'))
            icon_path = self.overlays_path/f"{token}.png"
            raw_icon = Image.open(icon_path).convert("RGBA")
            for i in range(value):
                region = side[f'{token}{i+1}_region']
                if not region:
                    continue
                icon = raw_icon.resize((region.width, region.height))
                card_image.paste(icon, (region.x, region.y), icon)

    def render_template(self, card_image, side):
        """Render a template image onto a card"""
        template_value = side['template']
        if '<class>' in side['template']:
            side_class = side.get('classes', ['guardian'])
            card_class = side_class[0] if len(side_class) == 1 else 'multi'
            template_value = template_value.replace('<class>', card_class)
        if Path(template_value).is_file():
            template_path = Path(template_value)
        else:
            template_path = self.templates_path / (template_value + '.png')

        try:
            template = Image.open(template_path).convert("RGBA")
            template = template.resize((card_image.width, card_image.height))
            card_image.paste(template, (0, 0), template)
        except Exception as e:
            print(f"Error rendering template: {str(e)}")

    def render_illustration(self, card_image, side):
        """Render the illustration/portrait"""
        # Get illustration path
        illustration_path = side.get('illustration', None)
        if not illustration_path or not os.path.exists(illustration_path):
            return

        try:
            # Load illustration
            illustration = Image.open(illustration_path).convert("RGBA")

            # Get clip region
            clip_region = Region(side['illustration_region'])

            rotation = side.get('illustration_rotation', 0)
            if rotation:
                illustration = illustration.rotate(rotation)

            # Calculate scaling
            illustration_scale = side.get('illustration_scale', None)
            if not illustration_scale:
                if illustration.width > illustration.height:
                    illustration_scale = clip_region.height / illustration.height
                else:
                    illustration_scale = clip_region.width / illustration.width

            # Resize illustration
            new_width = int(illustration.width * illustration_scale)
            new_height = int(illustration.height * illustration_scale)
            illustration = illustration.resize((new_width, new_height))

            # Apply panning
            pan_x = side.get('illustration_pan_x', 0)
            pan_y = side.get('illustration_pan_y', 0)

            # Position and paste
            card_image.paste(
                illustration,
                (clip_region.x+pan_x, clip_region.y+pan_y),
                illustration
            )
        except Exception as e:
            print(f"Error rendering illustration: {str(e)}")

    def render_rule_text(self, card_image, side):
        """Render the card rule text"""
        self.current_field = 'rule_text'
        rule_text = side.get('rule_text', '')
        if not rule_text:
            return

        text_region = Region(side['body_region'])

        # Checkboxes - primarily for Customizable
        value = side.get('checkbox_entries')
        if value:
            lines = ''
            for slots, name, text in value:
                lines += '\n' + '☐'*slots + f' <b>{name}.</b> {text}'
            rule_text += lines

        flavor_text = side.get('flavor_text', '')
        if flavor_text:
            rule_text += '\n'
            rule_text += f'<center><i>{flavor_text}</i>'

        # Get text polygon if available
        text_polygon = side.get('body_polygon', None)

        # Get alignment
        #alignment = self.get_setting('body_alignment', card_data, side, defaults, 'left')

        # Render using rich text renderer
        self.rich_text.render_text(
            card_image,
            rule_text,
            text_region,
            polygon=text_polygon,
            alignment='left'
        )

    def render_cost(self, card_image, side):
        """Render the card cost"""
        self.current_field = 'cost'
        cost = side.get('cost', None)
        if not cost:
            return

        cost_region = Region(side['cost_region'])

        # Draw cost
        if cost == '-':
            icon_path = self.assets_path/'numbers/AHLCG-Cost--.png'
            dash_icon = Image.open(icon_path).convert("RGBA")
            dash_2 = dash_icon.convert('HSV')
            h, s, v = dash_2.split()
            np_h = np.array(s, dtype=np.int8)
            np_h = (np_h * 0)
            h_shifted = Image.fromarray(np_h, 'L')
            new_img = Image.merge('HSV', (h, h_shifted, v))
            #new_img = new_img.resize((cost_region.width, cost_region.height))
            card_image.paste(new_img, (cost_region.x+5, cost_region.y+5), dash_icon)
            return

        self.rich_text.render_text(card_image, cost, cost_region, alignment='center', font="cost", font_size=40, outline=1, outline_fill='black', fill='white')

    def render_level(self, card_image, side):
        """Render the card level"""
        # Get level
        level = side.get('level')
        region = Region(side.get('level_region'))
        if not region:
            return

        if level is None or level == '':
            level_icon_path = self.overlays_path/f"no_level.png"
            level_icon = Image.open(level_icon_path).convert("RGBA")
            level_icon = level_icon.resize((level_icon.width*2, level_icon.height*2))
            card_image.paste(level_icon, (16, 7), level_icon)
            return

        level_icon_path = self.overlays_path/f"level_{level}.png"
        level_icon = Image.open(level_icon_path).convert("RGBA")
        level_icon = level_icon.resize((level_icon.width*2, level_icon.height*2))
        card_image.paste(level_icon, region.pos, level_icon)

    def render_class_symbols(self, card_image, side):
        """Render class symbols for multiclass cards"""
        classes = side.get('classes', [])
        if len(classes) < 2:
            return

        # Render each class symbol
        for index, cls in enumerate(classes):
            symbol_path = self.overlays_path/f"class_symbol_{cls}.png"
            region = Region(side.get(f"class_symbol_{index+1}_region"))

            symbol = Image.open(symbol_path).convert("RGBA")
            symbol = symbol.resize((symbol.width*2, symbol.height*2))
            card_image.paste(symbol, (region.x, region.y), symbol)

    def render_chaos(self, card_image, side):
        """ Renders the scenario reference cards.

            This could be handled in a lot of different ways,
            but this allows for easy json formatting of the card.
            It's essentially just a list of image and text.
        """
        entries = side.get('entries', [])
        if not entries:
            return

        region = Region(side['chaos_region'])
        font = side.get("chaos_font", {})
        for index, entry in enumerate(entries):
            tokens = entry['token']
            y = int(region.y + region.height/len(entries)*index)

            if type(tokens) != list:
                tokens = [tokens]

            for token_index, token in enumerate(tokens):
                try:
                    token_image = Image.open( overlay_dir / f"chaos_{token}.png")
                except:
                    continue
                token_image = token_image.resize((
                    int((region.width/4)/len(tokens)),
                    int((region.width/4)/len(tokens))
                ))
                x = int(region.x + token_image.width*token_index)
                card_image.paste(token_image, (x,y), token_image)

            self.rich_text.render_text(
                card_image,
                entry['text'],
                Region({'x': (region.x+region.width//3), 'y': y, 'height': region.height//len(entries), 'width': (2*region.width)//3}),
                font=font.get('font', 'regular'),
                font_size=font.get('size', 32),
                fill=font.get('color', '#231f20'),
                outline=font.get('outline'),
                outline_fill=font.get('outline_color'),
                alignment=font.get('alignment', 'left'),
            )

    def render_customizable(self, card_image, side):
        """ Renders the scenario reference cards.

            This could be handled in a lot of different ways,
            but this allows for easy json formatting of the card.
            It's essentially just a list of image and text.
        """
        if side['type'] != 'customizable':
            return

        entries = side.get('entries', [])
        if not entries:
            return

        region = Region(side['text_region'])
        font = side.get("text_font", {})
        parsed_text = ""
        for entry in entries:
            parsed_text += '\n' + int(entry[0])*'☐'
            parsed_text += f' <b>{entry[1]}.</b> '
            parsed_text += entry[2]

        self.rich_text.render_text(
            card_image,
            parsed_text,
            region,
            font=font.get('font', 'regular'),
            font_size=font.get('size', 32),
            fill=font.get('color', '#231f20'),
            outline=font.get('outline'),
            outline_fill=font.get('outline_color'),
            alignment=font.get('alignment', 'left'),
        )



from time import time

from kivy.config import Config
import sys

Config.set('input', 'mouse', 'mouse,disable_multitouch')
Config.set('kivy', 'log_enable', '0')
#Config.set('kivy', 'window_icon', 'assets/elder_sign_neon.ico')
Config.set('graphics', 'width', '1600')
Config.set('graphics', 'height', '900')

from kivy.app import App
import kivy_garden.contextmenu
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.splitter import Splitter
from kivy.uix.treeview import TreeView, TreeViewLabel
from kivy.uix.scrollview import ScrollView
from kivy.uix.label import Label
from kivy.uix.button import Button
from kivy.uix.textinput import TextInput
from kivy.uix.filechooser import FileChooserListView
from kivy.uix.popup import Popup
from kivy.uix.image import Image, CoreImage
from kivy.uix.behaviors import ButtonBehavior, FocusBehavior, DragBehavior
from kivy.properties import ObjectProperty, StringProperty, BooleanProperty, ListProperty
from kivy.clock import Clock
from kivy.core.window import Window
from kivy.base import ExceptionHandler, ExceptionManager

from shoggoth.editor import CardEditor, EncounterEditor, ProjectEditor, NewCardPopup
from shoggoth.project import Project

from kivy.logger import Logger, LOG_LEVELS
Logger.setLevel(LOG_LEVELS["info"])

import os
import json
from shoggoth.files import defaults_dir, asset_dir

from shoggoth.renderer import CardRenderer
from shoggoth.file_monitor import FileMonitor
from shoggoth.card import Card, TEMPLATES

from kivy.storage.jsonstore import JsonStore
from shoggoth.ui import show_file_select


class ShoggothRoot(FloatLayout):
    """Root widget for the Shoggoth application"""
    bg_path = StringProperty("")

    def __init__(self, **kwargs):
        super(ShoggothRoot, self).__init__(**kwargs)
        Window.bind(on_key_down=self.on_keyboard)
        bg_file = asset_dir/'parchment2.png'
        try:
            self.bg_path = str(bg_file)
        except Exception as e:
            print(f"Error loading background image: {e}")
        ExceptionManager.add_handler(Wing())

    def on_keyboard(self, instance, keyboard, keycode, text, modifiers):
        # Handle keyboard shortcuts
        if 'ctrl' in modifiers:
            if text == 's':
                f = App.get_running_app().save_changes
                Clock.schedule_once(lambda x: f())
                return True
            if text == 'n':
                f = App.get_running_app().open_new_card_dialog
                Clock.schedule_once(lambda x: f())
                return True
            if text == 'n':
                App.get_running_app().number_cards()
                return True
            if text == 'e':
                if 'shift' in modifiers:
                    f = App.get_running_app().export_all
                    Clock.schedule_once(lambda x: f())
                    return True
                else:
                    f = App.get_running_app().export_current
                    Clock.schedule_once(lambda x: f())
                    return True
            if text == 'o':
                f = App.get_running_app().open_project_dialog
                Clock.schedule_once(lambda x: f())
                return True
            if text == 'p':
                f = App.get_running_app().show_goto_dialog
                Clock.schedule_once(lambda x: f())
                return True
        return False

class Wing(ExceptionHandler):
    """ Exception handle, that carries off
        unhandled exceptions to a central location
    """
    def __init__(self):
        super().__init__()

    def handle_exception(self, exception):
        import traceback
        print("*** Exception handled by Wing ***")
        print(exception)
        print("*********************************")
        # Keyboard Interrupts should just kill the application.
        if type(exception) == KeyboardInterrupt:
            return ExceptionManager.RAISE
        # Everything else, is expected to be unhandled exceptions that
        # Should then pop up in front of the user
        App.get_running_app().show_ok_dialog(f'Something unexpected happened. This is a last effort to show you what happened. Maybe Shoggoth will crash after this.\n\n{exception}\n\n{traceback.format_exc(limit=-5)}')
        return ExceptionManager.PASS


class FileBrowser(BoxLayout):
    """File browser widget showing project files"""
    project = ObjectProperty()
    tree: TreeView

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.bind(project=self.refresh)

    def on_selected_node(self):
        app = App.get_running_app()
        if self.tree.selected_node.element_type == 'project':
            app.show_project(self.tree.selected_node.element)
        elif self.tree.selected_node.element_type == 'encounter':
            app.show_encounter(self.tree.selected_node.element)
        elif self.tree.selected_node.element_type == 'card':
            app.show_card(self.tree.selected_node.element)

    def refresh(self, *args):
        opens = set()
        for node in self.tree.iterate_open_nodes():
            opens.add((node.level, node.text))
        print('clearing tree, project is now', self.project)
        # Clear tree
        for node in self.tree.root.nodes:
            self.tree.remove_node(node)

        # Recursively add files and folder
        self._add_nodes()
        for node in self.tree.iterate_all_nodes():
            if (node.level, node.text) in opens:
                self.tree.toggle_node(node)

    def _add_nodes(self):
        # Add folder contents to tree
        p_node = self.tree.add_node(TreeViewButton(text=self.project['name'], element=self.project, element_type='project'))

        if self.project.encounter_sets and self.project.cards:
            campaign_node = self.tree.add_node(TreeViewButton(text='Campaign cards', element=None, element_type=''), p_node)
            player_node = self.tree.add_node(TreeViewButton(text='Player cards', element=None, element_type=''), p_node)
        else:
            campaign_node = player_node = p_node

        for encounter_set in self.project.encounter_sets:
            e_node = self.tree.add_node(TreeViewButton(text=encounter_set.name, element=encounter_set, element_type='encounter'), campaign_node)
            story_node = self.tree.add_node(TreeViewButton(text='Story', element=None, element_type=''), e_node)
            location_node = self.tree.add_node(TreeViewButton(text='Locations', element=None, element_type=''), e_node)
            encounter_node = self.tree.add_node(TreeViewButton(text='Encounter', element=None, element_type=''), e_node)
            for card in encounter_set.cards:
                if card.front.get('type') == 'location':
                    target_node = location_node
                elif card.back.get('type') == 'encounter':
                    target_node = encounter_node
                else:
                    target_node = story_node
                c_node = self.tree.add_node(TreeViewButton(text=card.name, element=card, element_type='card'), target_node)

        class_nodes = {}
        for cls in ('investigators', 'seeker', 'rogue', 'guardian', 'mystic', 'survivor', 'neutral', 'other'):
            class_nodes[cls] = self.tree.add_node(TreeViewButton(text=cls, element=None, element_type=''), player_node)

        investigator_nodes = {}

        for card in self.project.cards:
            if group := card.data.get('investigator', False):
                target_node = class_nodes.get('investigators')
                if not investigator_nodes.get(group):
                    investigator_nodes[group] = self.tree.add_node(TreeViewButton(text=group, element=None, element_type=''), target_node)
                target_node = investigator_nodes[group]
            else:
                target_node = class_nodes.get(card.get_class(), class_nodes['other'])
            c_node = self.tree.add_node(TreeViewButton(text=card.name, element=card, element_type='card'), target_node)

    def on_tree_select(self, instance, node):
        if hasattr(node, 'full_path') and node.full_path.endswith('.json'):
            self.parent.parent.load_card(node.full_path)


class TreeViewButton(TreeViewLabel):
    element = ObjectProperty()
    element_type = StringProperty('')


class CardPreview(BoxLayout):
    """Widget for displaying card previews"""
    def set_card_images(self, front_image, back_image):
        self.ids.front_preview.texture = front_image
        self.ids.back_preview.texture = back_image


class NewProjectPopup(Popup):
    """Popup for selecting files/folders"""
    def __init__(self, callback):
        super().__init__()
        self.callback = callback

    def save(self):
        data = Project.new(
            self.ids.name.text,
            self.ids.code.text,
            self.ids.icon.text,
        )
        with open(self.ids.file_name.text, 'w') as file:
            json.dump(data, file)

        self.dismiss()
        App.get_running_app().add_project_file(self.ids.file_name.text)

class AboutPopup(Popup):
    """ Information about the application """
    pass

class OkPopup(Popup):
    text = StringProperty("")
    """ Shows a text string, with no interaction """
    pass

class GotoEntryListItem(ButtonBehavior, BoxLayout):
    entry = ObjectProperty()
    callback = ObjectProperty()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

class GotoEntry:
    def __init__(self, item, type):
        self.id = str(item.id)
        self.type = type

        if type == 'card':
            if item.encounter:
                self.name = f'{item.expansion.name}/{item.encounter.name}/{item.name}'
            else:
                self.name = f'{item.expansion.name}/{item.name}'
        elif type == 'set':
            self.name = f'{item.expansion.name}/{item.name}'
        elif type == 'project':
            self.name = f'{item.name}'

class GotoPopup(Popup):
    """Popup for selecting files/folders"""
    def __init__(self):
        super().__init__()
        self.entries = []
        self.shown_entries = []
        self.get_all_entries()
        self.ids.input.bind(text=self.filter_entries)

    def get_all_entries(self):
        self.entries = []
        self.entries.extend([GotoEntry(entry, type='card') for entry in App.get_running_app().current_project.get_all_cards()])
        self.entries.extend([GotoEntry(entry, type='set') for entry in App.get_running_app().current_project.encounter_sets])
        self.entries.extend([GotoEntry(App.get_running_app().current_project, type='project')])

    def filter_entries(self, instance, text):
        self.shown_entries = [entry for entry in self.entries if text.lower() in entry.name.lower() or text.lower() in entry.id.lower()]
        self.ids.list.clear_widgets()
        for entry in self.shown_entries:
            self.ids.list.add_widget(GotoEntryListItem(entry=entry, callback=self.go))

    def go(self, entry):
        if entry.type == 'card':
            App.get_running_app().goto_card(entry.id)
        elif entry.type == 'set':
            App.get_running_app().goto_set(entry.id)
        elif entry.type == 'project':
            App.get_running_app().goto_project(entry.id)
        self.dismiss()

class ShoggothApp(App):
    """Main application class for Shoggoth Card Creator"""
    current_card_id = StringProperty("")
    current_project = ObjectProperty(None)
    status_message = StringProperty("Ready")

    def build(self):
        self.icon = str(asset_dir / 'elder_sign_neon.png')
        # Initialize main components
        self.storage = JsonStore('shoggoth.json')
        self.root = ShoggothRoot()

        # Initialize card renderer
        self.card_renderer = CardRenderer()
        self.scheduled_redraw = None

        # Initialize file monitoring
        self.file_monitor = FileMonitor(None, self.on_file_changed)

        # Initialize asset monitoring
        self.assets_monitor = FileMonitor(defaults_dir, self.on_file_changed)
        self.assets_monitor.start()

        # build a session object if one doesn't exist
        if not self.storage.exists('session'):
            print('no defaults, creating new storage')
            self.storage.put('session', project=None)
            self.storage.put('session', last_id=None)

        # restore the existing session
        if 'project' in self.storage.get('session'):
            self.current_project = Project.load(self.storage.get('session')['project'])
            self.root.ids.file_browser.project = self.current_project

        return self.root

    def goto_project(self, project_id):
        self.current_project = Project.load(project_id)
        self.root.ids.file_browser.project = self.current_project

    def goto_set(self, set_id):
        self.current_project = Project.load(self.storage.get('session')['project'])
        self.root.ids.file_browser.project = self.current_project

    def goto_card(self, card_id):
        print(f'looking for card with id {card_id}')
        tree = self.root.ids.file_browser.tree
        for node in tree.iterate_all_nodes():
            try:
                if str(node.element.id) == str(card_id):
                    tree.select_node(node)
                    parent = node.parent_node
                    while parent:
                        if not parent.is_open:
                            tree.toggle_node(parent)
                        parent = parent.parent_node
                    return
            except AttributeError:
                continue
        print('Card not found')

    def show_goto_dialog(self):
        if not self.current_project:
            return
        goto_dialog = GotoPopup()
        goto_dialog.open()
        goto_dialog.ids.input.focus = True

    def show_about_dialog(self):
        about_dialog = AboutPopup()
        about_dialog.open()

    def show_ok_dialog(self, text):
        dialog = OkPopup(text=text)
        dialog.open()

    def open_project_dialog(self):
        """Show file chooser dialog to select project folder"""
        show_file_select(None, callback=self.add_project_file)

    def open_new_card_dialog(self):
        dialog = NewCardPopup()
        dialog.open()

    def new_project_dialog(self):
        """Show popup to create new project"""
        creation_dialog = NewProjectPopup(
            callback=self.add_project_file
        )
        creation_dialog.open()

    def add_project_file(self, path):
        """Set the project path and initialize monitoring"""
        if not path:
            return

        project = Project.load(path)
        self.root.ids.file_browser.project = project
        self.current_project = project
        self.current_card = None
        self.storage.put('session', project=path)

        self.status_message = f"Project opened: {self.current_project.get('name')}"
        self.root.ids.status_bar.text = self.status_message

    def on_file_changed(self, file_path):
        """Handle file changes outside the editor"""
        self.load_file(self.current_card_path)

    def number_cards(self):
        self.current_card.expansion.assign_card_numbers()
        self.status_message = f"Updated card numbers"
        self.update_card_preview()

    def save_changes(self):
        self.current_project.save()

    def show_project(self, project):
        self.root.ids.editor_container.clear_widgets()
        self.root.ids.editor_container.add_widget(ProjectEditor(project=project))

    def show_encounter(self, encounter):
        self.root.ids.editor_container.clear_widgets()
        self.root.ids.editor_container.add_widget(EncounterEditor(encounter=encounter))

    def show_card(self, card):
        self.current_card = card
        self.current_card_id = card.id
        self.root.ids.editor_container.clear_widgets()
        self.root.ids.editor_container.add_widget(CardEditor(card=card))
        self.update_card_preview()

    def update_texture(self, texture, container):
        #tex = self.card_renderer.pil_to_texture(texture)
        img = Image(size_hint_y=None, height=300)
        container.add_widget(img)
        img.texture = CoreImage(texture, ext='jpeg').texture

    def render_thumbnail(self, card, container):
        texture = self.card_renderer.get_thumbnail(card)
        Clock.schedule_once(lambda x:self.update_texture(texture, container), 1)

    def update_card_data(self, face, field, value):
        print('update current card', face, field, value)
        if face.get(field, None) != value:
            self.current_card.set(face, field, value)

    def update_card_preview(self):
        print('update card preview', self.current_card)
        if self.scheduled_redraw:
            self.scheduled_redraw.cancel()
        self.scheduled_redraw = Clock.schedule_once(self._update_card_preview, 0.3)

    def _update_card_preview(self, *args, **kwargs):
        try:
            front_image, back_image = self.card_renderer.get_card_textures(self.current_card)
            self.root.ids.card_preview.set_card_images(front_image, back_image)
        except Exception as e:
            self.status_message = str(e)

    def refresh_tree(self):
        self.root.ids.file_browser.refresh()
        self.root.ids.file_browser.refresh()
        self.root.ids.file_browser.refresh()

    def load_project(self, path):
        """ Adds a project to the project explorer """
        if not path or not os.path.exists(path):
            return

        try:
            project = Project.load(path)

            # Update status
            self.status_message = f"Loaded: {os.path.basename(path)}"

        except Exception as e:
            print(e)
            self.status_message = f"Error loading project: {str(e)}"
            self.root.ids.status_bar.text = self.status_message

    def export_card(self, file_path=None):
        """Export the current card as an image"""
        if not self.card_data:
            self.status_message = "No card loaded to export"
            self.root.ids.status_bar.text = self.status_message
            return

        try:
            # Generate file path if not provided
            if not file_path:
                basename = os.path.basename(self.current_card_path)
                export_name = os.path.splitext(basename)[0] + ".png"
                file_path = os.path.join(os.path.dirname(self.current_card_path), export_name)

            # Get card images
            front_image, _ = self.card_renderer.get_card_images(self.card_data)

            # Save to file
            front_image.save(file_path, "PNG")

            self.status_message = f"Card exported to: {file_path}"
            self.root.ids.status_bar.text = self.status_message

        except Exception as e:
            self.status_message = f"Error exporting card: {str(e)}"
            self.root.ids.status_bar.text = self.status_message

    def on_stop(self):
        """Clean up when the application stops"""
        if self.file_monitor:
            self.file_monitor.stop()

if __name__ == "__main__":
    ShoggothApp().run()



from uuid import uuid4
from shoggoth.card import Card


class EncounterSet:
    def __init__(self, data, expansion=None):
        self.name = data['name']
        self.icon = data['icon']
        self.code = data.get('code', 'xx')
        self.data = data
        self.expansion = expansion
        self.get = data.get
        self.id = data.get('id', uuid4())
        self.__getitem__ = self.data.__getitem__

    def __eq__(self, other):
        return self.data == other.data and self.expansion == other.expansion

    @property
    def cards(self):
        for c in self.data['cards']:
            yield Card(c, encounter=self, expansion=self.expansion)

    @staticmethod
    def is_valid(data):
        return 'cards' in data and 'name' in data and 'icon' in data

    def add_card(self, card):
        if type(card) == Card:
            self.data['cards'].append(card.data)
        else:
            self.data['cards'].append(card)

    def assign_card_numbers(self):
        current_number = 1
        for card in self.cards:
            amount = card.data.get('amount', 2)
            if amount > 1:
                card.data['encounter_number'] = f'{current_number}-{current_number+amount-1}'
            else:
                card.data['encounter_number'] = f'{current_number}'
            current_number += amount
        self.data['card_amount'] = current_number-1



import argparse
import os
from shoggoth.files import asset_dir, root_dir
import urllib.request
import zipfile


def version_is_up_to_date() -> bool:
    """ Parses the asset pack version """
    if not (asset_dir/'version.txt').exists():
        return False
    with (asset_dir/'version.txt').open('r') as f:
        if not f.read().startswith('0.1.0'):
            return False
    return True

def run():
    os.environ["KIVY_NO_ARGS"] = "1"

    parser = argparse.ArgumentParser(description='Shoggoth Card Creator')
    parser.add_argument('-v', '--view', metavar='FILE', help='Open in viewer mode with specified file')
    args = parser.parse_args()

    # ensure directories exist
    root_dir.mkdir(parents=True, exist_ok=True)

    # ensure assets directory exists
    if not asset_dir.is_dir() or not version_is_up_to_date():
        print("Asset pack not found. Downloading assets...")
        # download assets
        url = 'https://www.dropbox.com/scl/fi/pp70yhzu7saqhhnd7xtb0/assets.zip?rlkey=ln16n1glarlb2z46af52mu2rd&st=vrwos3du&dl=1'
        filehandle, _ = urllib.request.urlretrieve(url)
        with zipfile.ZipFile(filehandle, 'r') as file:
            file.extractall(root_dir)
        print("Assets downloaded successfully.")

    if args.view:
        # Start in viewer mode
        from shoggoth.viewer import ViewerApp
        app = ViewerApp(args.view)
    else:
        # Start in normal mode
        from shoggoth.main import ShoggothApp
        app = ShoggothApp()
    app.run()



from kivy.uix.popup import Popup
from kivy.uix.dropdown import DropDown
from kivy.uix.button import Button
from kivy.app import App
from kivy.properties import ObjectProperty
from kivy_garden.contextmenu import ContextMenu, ContextMenuTextItem
import os
from shoggoth import card


class FileChooserPopup(Popup):
    """Popup for selecting files/folders"""
    def __init__(self, field, callback=None):
        super().__init__()
        self.field = field
        self.callback = callback
        self.ids.filechooser.path = os.path.expanduser("~/Documents")

    def select(self, path, file_name):
        self.dismiss()
        if self.callback:
            self.callback(os.path.join(path, file_name))
            return
        self.field.text = os.path.join(path, file_name)


class FileSavePopup(Popup):
    """Popup for selecting files/folders"""
    def __init__(self, field, callback=None):
        super().__init__()
        self.field = field
        self.callback = callback
        self.ids.filechooser.path = os.path.expanduser("~/Documents")

    def select(self, path, file_name):
        self.dismiss()
        if self.callback:
            self.callback(os.path.join(path, file_name))
            return
        self.field.text = os.path.join(path, file_name)

class ClassSelectDropdown(DropDown):
    def __init__(self, callback):
        super().__init__()
        self.callback = callback

    def select_value(self, value):
        self.callback(value)

class ValueSelectPopup(Popup):
    def __init__(self, callback, **kwargs):
        super().__init__(**kwargs)
        self.callback = callback
        self.populate()

    def populate(self):
        """ empty method for populating data fields """
        pass

    def select(self, value):
        self.dismiss()
        self.callback(value)


class TemplateSelector(ValueSelectPopup):
    pass


class SetSelector(ValueSelectPopup):
    def populate(self):
        self.ids.option_container.add_widget(
            Button(
                text='Player card',
                on_release=lambda n: self.select(None),
                size_hint_y=None,
                height='32dp',
            )
        )
        project = App.get_running_app().current_project
        for encounter in project.encounter_sets:
            def action(*args, encounter=encounter, **kwargs):
                self.select(encounter)
            self.ids.option_container.add_widget(
                Button(
                    text=encounter.name,
                    on_release=action,
                    size_hint_y=None,
                    height='32dp',
                )
            )



class TypeSelectDropdown(ContextMenu):
    target = ObjectProperty()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.create_values()

    def select(self, widget):
        print('select with', widget)
        self.target.input.text = widget.text
        self.hide()

    def create_values(self):
        from shoggoth import files

        for file in files.defaults_dir.iterdir():
            self.add_widget(
                ContextMenuTextItem(
                    text=file.stem,
                    on_release=self.select,
                )
            )


def show_file_select(target, callback=None):
    pop = FileChooserPopup(target, callback=callback)
    pop.open()

def show_file_save(target, callback=None):
    pop = FileSavePopup(target, callback=callback)
    pop.open()

def show_class_select(parent, callback):
    dropdown = ClassSelectDropdown(callback=callback)
    dropdown.open(parent)



import os
import json
from uuid import uuid4

from kivy.app import App
from shoggoth.card import Card
from shoggoth.encounter_set import EncounterSet

type_order = {
    "scenario": 0,
    "chaos": 1,
    "agenda": 2,
    "act": 3,
    "location": 4,
    "story": 5,
    "key": 6,
    "treachery": 7,
    "enemy": 8,
    "story": 9,
    "investigator": 10,
    "other": 11
}

class_order = {
    "guardian": 0,
    "seeker": 1,
    "rogue": 2,
    "mystic": 3,
    "survivor": 4,
    "neutral": 5,
    "multi": 6  # Multi-class cards are sorted last
}
def sort_cards(cards):
    cards.sort(key=lambda card: (
        type_order.get(card.front['type'], type_order["other"]),
        card.front.get('agenda_index', -1),
        card.front.get('act_index', -1),
        class_order.get(card.front.get('class'), -1),
        card.front.get('level', -1),
        card.front.get('type', ''),
        card.name,
    ))

class Project:
    """ Class to handle project files

        Projects are ultimately just representations of json files.
    """
    def __init__(self, file_path, data):
        self.file_path = file_path
        self.data = data
        self.icon = data.get('icon', '')
        self.code = data.get('code', 'xx')
        self.id = data.get('id', uuid4())

    def __eq__(self, other):
        return self.data == other.data

    def __getitem__(self, key):
        return self.data[key]

    def get(self, key, default=None):
        return self.data.get(key, default)

    @property
    def name(self):
        return self.data['name']

    @property
    def cards(self):
        for card in self.data.get('cards', []):
            yield Card(card, expansion=self)

    @property
    def encounter_sets(self):
        # order sets to always come out right
        self.data['encounter_sets'].sort(key=lambda x: (
            x.get('order', '999'),
            x.get('name'),
        ))

        for e in self.data['encounter_sets']:
            yield EncounterSet(e, expansion=self)

    def assign_card_numbers(self):
        current_number = 1
        for encounter_set in self.encounter_sets:
            encounter_set.assign_card_numbers()
            for card in encounter_set.cards:
                card.expansion_number = current_number
                current_number += 1
        for card in self.cards:
            card.expansion_number = current_number
            current_number += 1

    def add_card(self, card):
        if not 'cards' in self.data:
            self.data['cards'] = []
        if type(card) == Card:
            self.data['cards'].append(card.data)
        else:
            self.data['cards'].append(card)

    def get_all_cards(self):
        result = []
        for e in self.encounter_sets:
            for c in e.cards:
                result.append(c)
        if 'cards' in self.data:
            for c in self.data['cards']:
                result.append(Card(c, expansion=self))
        return result

    @staticmethod
    def load(file_path):
        """Load card data from JSON file"""
        try:
            with open(file_path, 'r') as f:
                data = json.load(f)
        except Exception as e:
            print(e)
            raise

        for entry in data["encounter_sets"]:
            try:
                assert EncounterSet.is_valid(entry)
            except AssertionError as e:
                print('Entry failed assertion, project is invalid: ', entry)
                raise Exception('Invalid project file.')

        return Project(file_path, data)

    def add_encounter_set(self, name):
        if 'encounter_sets' not in self.data:
            self.data['encounter_sets'] = []
        for enc in self.data['encounter_sets']:
            if enc['name'] == name:
                raise Exception('duplicate encounter set name')

        encounter_data = {
            'name': name,
            'icon': '',
            'cards': [],
        }
        self.data['encounter_sets'].append(encounter_data)
        App.get_running_app().refresh_tree()

    def remove_encounter_set(self, index):
        self.data['encounter_sets'].pop(index)

    def save(self):
        """Save data to file"""
        with open(self.file_path, 'w') as f:
            json.dump(self.data, f, indent=4)

    @staticmethod
    def new(name, code, icon):
        """ Returns the base data for a new Project """
        return {
            'name': name,
            'code': code,
            "icon": icon,
            'encounter_sets': [],
            'cards': [],
        }
